//*****************************************************************//
//                                                                 //
//  Huffman Algorithm THuffNode Object                             //
//  Copyright BrandsPatch LLC                                     //
//  http://www.explainth.at                                        //
//                                                                 //
//  All Rights Reserved                                            //
//                                                                 //
//  Permission is granted to use, modify and redistribute          //
//  the code in this Delphi unit on the condition that this        //
//  notice is retained unchanged.                                  //
//                                                                 //
//  BrandsPatch declines all responsibility for any losses,        //
//  direct or indirect, that may arise  as a result of using       //
//  this code.                                                     //
//                                                                 //
//                                                                 //
//*****************************************************************//
unit HuffHelp;

interface

uses Windows,SysUtils,Classes,BitList;

type TCodeBytes = array[1..256] of Byte;
     THuffBytes = array[0..255] of Byte;
     THuffCounts = array[0..255] of Integer;
     TSizes = array[0..255] of Byte;
     TShiftMask = array[1..8] of Byte;
     TCompBytes = array[0..MAXINT - 1] of Byte;
     PCompBytes = ^TCompBytes;

const LeftMasks:TShiftMask = ($01,$03,$07,$0F,$1F,$3F,$7F,$FF);

type THuffChars = class(TObject)
private
  FCount:Byte;
  FPosition:Word;
  FChars:array[0..255] of Byte;
public
  property CharCount:Byte read FCount;
  procedure WriteChar(AChar:Byte);
end;

type TWriteCodeProc = procedure of Object;

type THuffNode = class(TObject)
private
  FChar:Byte;
  //what character does this node represent - nil if a root node
  FFlag:Boolean;
  //does this node have a Huffman code more than 8 bits in length
  FFull:Byte;
  //if so, how many full bytes does the code require
  FPart:Byte;
  //how many additional bits, a partial byte, are then needed
  FSize:Word;
  FCount:Integer;
  //what is the frequency of FChar (if a leaf node)
  FCode:array[0..31] of Byte;
  {the Huffman code that represents this node. Can be between 0 and 256 bits}
  FNodes:array[0..1] of THuffNode;
  //if a root node, FNodes will be the two owned leaves.
  FWriters:array[Boolean] of TWriteCodeProc;
  {How we write the Huffman code depends on whether it is <= or >
   8 bits long. When writing out the compressed data this will be
   done very often. Not much point testing code length each time
   so we store the two writers - <=8Writer and >8Writer - in
   FWriters and then dereference the correct one with the aid
   of the FFlags boolean}
  procedure DesignateWriters;//populates FWriters
  procedure WritePolyCode;//<=8Writer
  procedure WriteUniCode;//>8Writer
  //Helpers
public
  constructor CreateAsLeaf(AChar,ACount:Integer);
  constructor CreateAsRoot(ANodes:array of THuffNode);
  constructor CreateFromStream(AStream:TStream;ABits:TBitList);
  destructor Destroy;override;
  procedure WriteToStream(AStream:TStream;ABits:TBitList);

  function AddIfMatch(AChar:Integer;AList:TList):Boolean;
  procedure BuildCode(ACode:TCodeBytes;ASize:Integer);
  function LeafReached(ACode:TCodeBytes;var ARead:Integer):Boolean;
  function Frequency:Integer;
  procedure WriteWithWriter;
end;

var _Left,_Right,_Diff:Integer;
    _Byte,_Bit,_Bi:Integer;
    _PD,_PS:PCompBytes;
    _BBFlag:Boolean;

implementation
//--------------------------------------------------------------------THuffChars
procedure THuffChars.WriteChar(AChar:Byte);
begin
  FChars[FPosition]:=AChar;
  inc(FPosition);
end;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++THuffNode Destroy
constructor THuffNode.CreateFromStream(AStream:TStream;ABits:TBitList);
begin
  inherited Create;
  case ABits.FetchBit of
   False:begin
            FNodes[0]:=THuffNode.CreateFromStream(AStream,ABits);
            FNodes[1]:=THuffNode.CreateFromStream(AStream,ABits);
          end;
   True:AStream.ReadBuffer(FChar,sizeof(FChar));
  end;
end;

destructor THuffNode.Destroy;
begin
  FNodes[0].Free;
  FNodes[1].Free;
  inherited;
end;

procedure THuffNode.WriteToStream(AStream:TStream;ABits:TBitList);

  procedure WriteLeaf;
  begin
    ABits.AddBit(True);
    AStream.WriteBuffer(FChar,sizeof(FChar));
  end;

  procedure WriteNonLeaf;
  begin
    ABits.AddBit(False);
    FNodes[0].WriteToStream(AStream,ABits);
    FNodes[1].WriteToStream(AStream,ABits);
  end;

begin
  if (FNodes[0] <> nil) then WriteNonLeaf else WriteLeaf;
end;
//------------------------------------------------------------THuffNode Builders
constructor THuffNode.CreateAsLeaf(AChar,ACount:Integer);
begin
  inherited Create;
  FChar:=AChar;
  FCount:=ACount;
  DesignateWriters;
  {Each leaf node represents a character, FChar, and has a
   certain frequency, FCount. It writes its Huffman code
   out using a procedure that depends on whether the code is
   <= or > 8 bits (1 byte) long. A leaf node has no children}
end;

constructor THuffNode.CreateAsRoot(ANodes:array of THuffNode);
var i:Integer;
begin
  inherited Create;
  for i:=0 to 1 do FNodes[i]:=ANodes[i];
  DesignateWriters;
  {A root node always has two children - two roots, two leaves or a
   root and a leaf}
end;

procedure THuffNode.DesignateWriters;
begin
  FWriters[False]:=WriteUniCode;
  FWriters[True]:=WritePolyCode;
  {The way we write out a node depends on whether it occupies
   more or less than 8 bits. }
end;

procedure THuffNode.WriteWithWriter;
begin
  FWriters[FFlag];
end;

function THuffNode.AddIfMatch(AChar:Integer;AList:TList):Boolean;
begin
  if (FNodes[0] = nil) then
  begin
    Result:=(FChar = AChar);
    if Result then AList.Add(Self);
  end else Result:=FNodes[0].AddIfMatch(AChar,AList) or FNodes[1].AddIfMatch(AChar,AList);
  {Although Huffman compression requires us to put nodes into a binary tree, when
   we want to reference the nodes it is a great deal faster to access them from a list.
   This is what is happening here - after building the tree we check for each leaf
   node (some may not exist because there is no corresponding value in the original
   data) we put the node back into the list. Its position in the list matches the
   byte value it encodes using Huffman.

   So we check to see if the node is a leaf - no children so FNodes[0] is nil - and
   if the character it manages is a match. If so, we add it to the list. Otherwise
   it is a non-leaf so we run the same test with each of its children}
end;

procedure THuffNode.WritePolyCode;
begin
  _Bi:=0;
  repeat
    _Left:=8 - _Bit;
    _PD^[_Byte]:=_PD^[_Byte] or ((FCode[_Bi] and LeftMasks[_Left]) shl _Bit);

    inc(_Byte);
    _PD^[_Byte]:=_PD^[_Byte] or (FCode[_Bi] shr _Left);
    inc(_Bi);
  until (_Bi = FFull);

  {$DEFINE _PolyCode}
  if (FPart > 0) then {$I WriteUniCode.INC}
  {$UNDEF _PolyCode}//  {$I WritePolyCode.INC}
end;

procedure THuffNode.WriteUniCode;
begin
  {$I WriteUniCode.INC}
end;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++THuffNode Rebuildes
function THuffNode.LeafReached(ACode:TCodeBytes;var ARead:Integer):Boolean;
begin
  Result:=(FNodes[0] = nil);
  if Result then
  begin
    _PD^[_Bi]:=FChar;
    inc(_Bi);
    {if we have reached a leaf node we know our output char so
     write it and move _Bi to point to the next element of _PD^}
  end else
  begin
    _Left:=ACode[ARead];
    {_Left will be either 0, meaning move left down the tree, or
    1 meaning move right down the tree. If we hit $FF we have run
    out of move instructions because the code we have read back
    is incomplete. So we give up, fetch one more bit and try again.

    There is an improvement possible here - as things stand, the
    next time round we restart the leaf search from the root. This
    is somewhat wasteful since we know that the leaf must lie below
    the node at which the current failure, _Left = $FF, occurred.
    The improvement would require two things

    a. Return the failure node and start the next leaf search there
       rather than from FRoot
    b. In the Inflate routine ARead will have to start not from 1
       but from ARead - 1 in this routine.

       Try it if you like. There is a tradeoff here - with a large
       data set containing very long Huffman codes this MIGHT
       result in efficiency. However, the extra logic and parameters
       required might well nullify the benefits - especially when
       you consider that the most frequent Huffman codes will infact
       be quite short.
    }
    inc(ARead);
    Result:=(_Left <> $FF) and FNodes[_Left].LeafReached(ACode,ARead);
  end;
end;
//-------------------------------------------------------------THuffNode Helpers
procedure THuffNode.BuildCode(ACode:TCodeBytes;ASize:Integer);
var i,j,AByte,ABit,ACount:Integer;
begin
  {ACode is an array of 256 bytes. For starters it contains all
   zeros. As we move through the tree, some of its values become
   1s - for each right hand child node.}
  FSize:=ASize;
  FFull:=ASize div 8;
  //how many whole bytes do we need to record this Huffman code
  FPart:=ASize mod 8;
  //how many bits -a parital byte - do we need in addition
  ACount:=FFull shl 3 - 1;j:=1;
  //how many fullbyte bits do we need to write
  FFlag:=(FFull > 0);

  for i:=0 to ACount do
  begin
    AByte:=i div 8;//which byte of FCode do we want to write to?
    ABit:=i mod 8;//and which bit of that byte?
    FCode[AByte]:=FCode[AByte] or (ACode[j] shl ABit);
    inc(j);
  end;
  {The most frequent byte patterns will have fractional byte
  Huffman codes - FFull = 0 - so this loop will not be executed}

  ABit:=0;
  for i:=j to ASize do
  begin
    FCode[FFull]:=FCode[FFull] or (ACode[i] shl ABit);
    inc(ABit);
  end;
  {FCode now contains the bit sequence corresponding to the
   correct Huffman code for FChar}

  if (FNodes[0] <> nil) then
  begin
    inc(ASize);
    //child nodes will have a Huffman code that is 1 bit longer
    //ACode for the left hand child node is zero, so we change nothing
    FNodes[0].BuildCode(ACode,ASize);
    ACode[ASize]:=1;
    //the right hand child writes a 1 so we set ACode[ASize] to 1
    FNodes[1].BuildCode(ACode,ASize);
  end;
end;

function THuffNode.Frequency:Integer;
begin
  if (FNodes[0] = nil) then Result:=FCount else Result:=(FNodes[0].Frequency + FNodes[1].Frequency);
  //if a child node, FNodes[0] = nil, just return FCount
  //otherwise return teh sum of the frequencies of the children
end;


end.
