//*****************************************************************//
//                                                                 //
//  Huffman Algorithm Impelmentation                               //
//  Copyright BrandsPatch LLC                                     //
//  http://www.explainth.at                                        //
//                                                                 //
//  All Rights Reserved                                            //
//                                                                 //
//  Permission is granted to use, modify and redistribute          //
//  the code in this Delphi unit on the condition that this        //
//  notice is retained unchanged.                                  //
//                                                                 //
//  BrandsPatch declines all responsibility for any losses,        //
//  direct or indirect, that may arise  as a result of using       //
//  this code.                                                     //
//                                                                 //
//                                                                 //
//*****************************************************************//
unit Huffman;

interface

uses Windows,SysUtils,Classes,HuffHelp,BitList;

type TDuo = record
       word: byte;
       freq: integer;
     end;

type THuffman = class(TObject)
private
  FBits:TBitList;
  FList:TList;
  FChars:THuffChars;
  FCounts:THuffCounts;
  FRoot:THuffNode;

  procedure BuildLeafNodes;
  procedure BuildNodeTree;
  procedure ConstructCodes;
  function ReBuildTree(AStream:TMemoryStream):Integer;
  function ReBuildTreeRFD(AStream: TMemoryStream):Integer;
  function TreeStream:TMemoryStream;
public
  destructor Destroy;override;

  function Compress(ASource:TMemoryStream):TMemoryStream;
  function DeCompress(ASource:TMemoryStream):TMemoryStream;
  function DeCompressRFD(ASource:TMemoryStream):TMemoryStream;
end;

implementation

//.....................................................THuffman Create & Destroy
function BuildByteTable(AStream:TMemoryStream):THuffCounts;
var i,ASize:Integer;
    P:PCompBytes;
begin
  with AStream do
  begin
    P:=Memory;
    ASize:=Size;
  end;

  FillChar(Result,sizeof(Result),0);
  for i:=0 to ASize - 1 do inc(Result[P^[i]]);
  {Here we build a frequency table - how many of each kind of,
  i.e. 0 to 255, byte values in the original data stream}
end;

function ReadBit:Integer;
begin
  Result:=ord(_PS^[_Byte] and (1 shl _Bit) > 0);
  //returns 1 if the bit is set
  inc(_Bit);
  inc(_Byte,ord(_Bit = 8));
  //when 8 bits have been read we move to the next byte
  _Bit:=_Bit*ord(_Bit < 8);
  {when Bit 8 has been read & we have moved to the next byte we
   restart from Bit 0 in the new byte}
end;

destructor THuffman.Destroy;
begin
  FBits.Free;
  FList.Free;
  FChars.Free;
  FRoot.Free;
  inherited;
end;
//******************************************************THuffman Precompresssion
procedure THuffman.BuildLeafNodes;
var i:Integer;
begin
  FreeAndNil(FRoot);
  with FList do
  begin
    Clear;
    for i:=0 to 255 do
    if (FCounts[i] > 0) then Add(THuffNode.CreateAsLeaf(i,FCounts[i]));
  end;
end;

procedure THuffman.BuildNodeTree;
var ANode,BNode:THuffNode;

  function LeastfrequentNode:THuffNode;
  var i,j,AFreq,AMin:Integer;
  begin
    AMin:=MAXINT;
    with FList do
    try
      for i:=0 to Count - 1 do with THuffNode(Items[i]) do
      begin
        AFreq:=Frequency;
        if (AFreq < AMin) then
        begin
          j:=i;
          AMin:=AFreq;
        end;
      end;
    finally
      Result:=Items[j];
      Delete(j);
    end;
    {We locate the leaf node with the lowest frequency still left
    in FList and return it as the result - it gets taken out of
    FList}
  end;

  function LocateNodes:Boolean;
  begin
    Result:=(FList.Count > 1);
    if Result then
    begin
      ANode:=LeastFrequentNode;
      BNode:=LeastFrequentNode;
      {ANode.Frequency <= BNode.Frequency!
       As things stand our tree is being constructed with
       rhsChild.Frequency >= lhsChild.Frequency. This means
       that we end up writing more 1s than 0s. It may be worth
       investigating whether the converse gives a better compression
       speed. In theory it should - writing 0s is faster than writing 1s
      }
    end;
  end;

begin
  with FList do while LocateNodes do Add(THuffNode.CreateAsRoot([ANode,BNode]));
  {We systematically take nodes out of the raw list and make them the
   children of a root node. At each step FList contains fewer leaves
   and more roots. In the end there is only one entry in FList. We }
  with FList do try FRoot:=Items[0] finally Delete(0) end;
  //FList is now empty
end;

procedure THuffman.ConstructCodes;
var i:Integer;
    ACode:TCodeBytes;
begin
  FillChar(ACode,sizeof(ACode),0);
  with FList,FRoot,FChars do
  begin
    BuildCode(ACode,0);
    {Having got the Huffman binary tree in place we proceed to
     construct the Huffman codes for each leaf node in that tree.
     For each leftward move down the tree we add a 0 and for each
     rightward move we add a 1.}
    for i:=0 to 255 do
    if AddIfMatch(i,FList) then WriteChar(i) else Add(nil);
    {To write out the compressed data we use the FCode data stored
     in each leaf node in the BuildCode step. This will happen over
     and over again. We don't want to have to negotiate the tree
     each time to locate the right leaf node. So we put all the
     leaves back into FList.

     This is just for convenience. The leaves are owned by their
     respective roots and the ultimate owner is FRoot. FList owns
     nothing.

    There may not be characters corresponding to all byte values
    in the original data. Nevertheless, it is convenient to create
    a list of leaf nodes with holes - nils where there are no leaves
    FList now contains nodes for #i in position i - could be nil if no #i in data
    FList DOES NOT own these nodes - they are owned, ultimately, by FRoot}
  end;
end;
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++Tree Streaming
function THuffMan.ReBuildTree(AStream:TMemoryStream):Integer;
begin
  FList:=TList.Create;
  FChars:=THuffChars.Create;
  FBits:=TBitList.CreateFromStream(AStream);
  FRoot:=THuffNode.CreateFromStream(AStream,FBits);
  ConstructCodes;
  with AStream do
  begin
    ReadBuffer(Result,sizeof(Result));//size of uncompressed data
    _PS:=Memory;
    _Byte:=Position;
  end;
end;

function THuffMan.ReBuildTreeRFD(AStream: TMemoryStream):Integer;
var duolist: array[1..256] of TDuo;
  curFreq: integer;
  numDuo, x: integer;
  info: array[0..1] of byte;
  NumEntries: Byte;
  Huffman: THuffman;
  CStream, DStream: TMemoryStream;
  FreqList: TList;
begin

  FList:=TList.Create;
  FChars:=THuffChars.Create;

  CurFreq := 0;
  NumDuo := 0;
  AStream.ReadBuffer(Result,4);
  AStream.ReadBuffer(NumEntries,1);
  AStream.ReadBuffer(CurFreq,1);
  AStream.Seek(1032,soFromBeginning);

  for x := 1 to 512 do
  begin
    AStream.ReadBuffer(info,2);
    if info[1] = 1 then
      inc(curFreq)
    else
    begin
      inc(numDuo);
      duolist[numDuo].word := info[0];
      duolist[numDuo].freq := curfreq;
    end;

    if curFreq > numEntries then
      break;

  end;

  for x := 1 to numDuo do
    FList.Add(THuffNode.CreateAsLeaf(duoList[x].word,duoList[x].freq));

  FreeAndNil(FRoot);
//  BuildLeafNodes;
  {then create a raw list of leaf nodes, one leaf for each
     possible byte value.}
  //allow for an empty memory stream
  if FList.Count > 0 then
  begin
    BuildNodeTree;//from the raw list build a node tree.
    ConstructCodes;//now build the huffman codes
  end;

  AStream.Seek(2048+8,soFromBeginning);
  
  with AStream do
  begin
    _PS:=Memory;
    _Byte:=Position;
  end;
end;

function THuffman.TreeStream:TMemoryStream;
var AStream:TMemoryStream;
begin
  AStream:=TMemoryStream.Create;
  try
    FRoot.WriteToStream(AStream,FBits);
    Result:=TMemoryStream.Create;
    FBits.WriteToStream(Result);
    Result.CopyFrom(AStream,0);
  finally AStream.Free end;
end;
//--------------------------------------------------THuffman Compress & DeCompress
function THuffman.Compress(ASource:TMemoryStream):TMemoryStream;
var i,ASize:Integer;
    ADest:TMemoryStream;

  function Prepare:Boolean;
  begin
    FBits:=TBitList.CreateEx;
    FList:=TList.Create;
    FChars:=THuffChars.Create;
    //characters & their count are stored here

    FCounts:=BuildByteTable(ASource);
    //first build the byte frequency table
    BuildLeafNodes;
    {then create a raw list of leaf nodes, one leaf for each
     possible byte value.}
    Result:=(FList.Count > 0);
    //allow for an empty memory stream
    if Result then
    begin
      BuildNodeTree;//from the raw list build a node tree.
      ConstructCodes;//now build the huffman codes
    end;
  end;

begin
  if Prepare then
  try
    ADest:=TMemoryStream.Create;
    ASize:=ASource.Size;//ASize bytes in the original data
    ADest.SetSize(ASize shl 1);
    {output will be smaller but we don't know its size so give it a
    safe initial size}
    _PD:=ADest.Memory;
    FillChar(_PD^,ASize shl 1,0);
    _PS:=ASource.Memory;
    {Accessing the memory locations for the input and output data is a very
     frequent task during compression. To make it faster grab pointers to
     the starts of the two memory locations and store them in globals}

    _Byte:=0;_Bit:=0;
    //ADest is empty so we start writing at Byte 0, Bit 0
    with FList do for i:=0 to ASize - 1 do THuffNode(Items[_PS^[i]]).WriteWithWriter;
    {For each byte of original data we must output a Huffman code. _PS^[i] derefs
    that byte. FList[0..255] contains the corresponding Huffman node with the
    right code. How the code is written depends on whether it is a <=8bit code
    or a >8bit code. WriteWithWriter handles that}

    inc(_Byte);//we start at 0 so size is +1!
    ADest.SetSize(_Byte);
    Result:=TreeStream;
    with Result do
    begin
      WriteBuffer(ASize,sizeof(ASize));
      WriteBuffer(_PD^,_Byte);
      Position:=0;
    end;
  finally ADest.Free end else Result:=nil;
end;

function THuffman.DeCompress(ASource:TMemoryStream):TMemoryStream;
var ASize:Integer;

  procedure Inflate(ALast:Integer);
  var i,ARead:Integer;
      ACode:TCodeBytes;
  begin
    repeat
      i:=0;
      repeat
        inc(i);ARead:=1;
        ACode[i]:=ReadBit;
        ACode[i + 1]:=$FF;
        {We keep reading bits from the compressed data stream until
         we find that we have a bit pattern matching a leaf node.

         As we read the data we are building up in the ACodes
         array byte values of 0 or 1 corresponding to the
         Huffman code of the character to be matched. Needless to
         say the match could fail because we haven't yet read a
         full code. We use the $FF value as a guard byte - when
         searching for a leaf if we find that we reach $FF in
         ACodes we know that we should give up, fetch one more
         bit and try again

         The process of finding that leaf node starts with the
         root. At each node we check to see if it has no children,
         i.e FNodes[0] = nil. If so it is a leaf node. The tree
         traversal works thus - each bit value we read tells us
         whether we should look in the left hand, FNodes[0], or
         right hand, FNodes[1], child of the current node.
        }
      until FRoot.LeafReached(ACode,ARead);
    until (_Bi > ALast);
    //stop once all characters in original data are regenerated
  end;

begin
  ASize:=ReBuildTree(ASource);

  Result:=TMemoryStream.Create;
  with Result do
  begin
    SetSize(ASize + 1);
    _PD:=Memory;
    {We grab a pointer to the memory block to which we will be
     outputting decompressed data. PD is a pointer to an array
     of bytes. While writing successive bytes we use this array
     indexed using the global _Bi. This is faster than trying
     to work with TMemoryStream methods - which in any case
     would not provide access to individual bytes of memory
     for bitwise manipulation}
  end;

  _Bit:=0;_Bi:=0;
  {The first bit we fetch is zero and once we know it we write it to
  Byte, _Bi, 0 of the PD^ array}
  try
    Inflate(ASize - 1);
    //we want to rebuild ASize - 1 characters, starting from the 0th!
  finally
    Result.SetSize(ASize);
    Result.Position:=0;
  end;
end;

function THuffman.DeCompressRFD(ASource:TMemoryStream):TMemoryStream;
var ASize:Integer;

  procedure Inflate(ALast:Integer);
  var i,ARead:Integer;
      ACode:TCodeBytes;
  begin
    repeat
      i:=0;
      repeat
        inc(i);ARead:=1;
        ACode[i]:=ReadBit;
        ACode[i + 1]:=$FF;
        {We keep reading bits from the compressed data stream until
         we find that we have a bit pattern matching a leaf node.

         As we read the data we are building up in the ACodes
         array byte values of 0 or 1 corresponding to the
         Huffman code of the character to be matched. Needless to
         say the match could fail because we haven't yet read a
         full code. We use the $FF value as a guard byte - when
         searching for a leaf if we find that we reach $FF in
         ACodes we know that we should give up, fetch one more
         bit and try again

         The process of finding that leaf node starts with the
         root. At each node we check to see if it has no children,
         i.e FNodes[0] = nil. If so it is a leaf node. The tree
         traversal works thus - each bit value we read tells us
         whether we should look in the left hand, FNodes[0], or
         right hand, FNodes[1], child of the current node.
        }
      until FRoot.LeafReached(ACode,ARead);
    until (_Bi > ALast);
    //stop once all characters in original data are regenerated
  end;

begin
  ASize:=ReBuildTreeRFD(ASource);

  Result:=TMemoryStream.Create;
  with Result do
  begin
    SetSize(ASize + 1);
    _PD:=Memory;
    {We grab a pointer to the memory block to which we will be
     outputting decompressed data. PD is a pointer to an array
     of bytes. While writing successive bytes we use this array
     indexed using the global _Bi. This is faster than trying
     to work with TMemoryStream methods - which in any case
     would not provide access to individual bytes of memory
     for bitwise manipulation}
  end;

  _Bit:=0;_Bi:=0;
  {The first bit we fetch is zero and once we know it we write it to
  Byte, _Bi, 0 of the PD^ array}
  try
    Inflate(ASize - 1);
    //we want to rebuild ASize - 1 characters, starting from the 0th!
  finally
    Result.SetSize(ASize);
    Result.Position:=0;
  end;
end;

end.
